#|

 Gtk ffi

 Copyright (c) 2004 by Vasilis Margioulas <vasilism@sch.gr>

 You have the right to distribute and use this software as governed by 
 the terms of the Lisp Lesser GNU Public License (LLGPL):

    (http://opensource.franz.com/preamble.html)
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 Lisp Lesser GNU Public License for more details.
 
|#

;;; Function with equivalents in gtklib.

(in-package :gtk-ffi)

(defun gtk-signal-connect (widget signal fun &key (after t) data destroy-data)
  #+shhtk (print (list "passing fun to gtk-signal-connect" signal fun))
  (g-signal-connect-data widget signal fun data destroy-data after))

(defun g-signal-connect-data (self detailed-signal c-handler data destroy-data after)
  (with-cstrings ((c-detailed-signal detailed-signal))
    (let ((p4 (or data c-null)))
      (g_signal_connect_data
       self
       c-detailed-signal
       (wrap-func c-handler)
       p4
       (or destroy-data c-null)
       (if after 1 0)))))

(uffi:def-function ("g_signal_connect_data" g_signal_connect_data)
    ((instance :pointer-void) (detailed-signal :cstring)
     (c-handler :pointer-void) (data :pointer-void)(destroy-data :pointer-void) (after :int))
  :returning :unsigned-long :call-direct nil)

(defun wrap-func (func-address) ;; vestigial. func would never be nil. i think.
  (or func-address 0))


(defun gtk-signal-connect-swap (widget signal fun &key (after t) data destroy-data)
  (g-signal-connect-closure widget signal
    (g-cclosure-new-swap (wrap-func fun) data destroy-data) after))

(defun gtk-object-set-property (obj property val-type val)
  (with-g-value (value)
      (g-value-init value (value-type-as-int val-type))
      (funcall (value-set-function val-type)
        value val)

      (g-object-set-property obj property value)

      (g-value-unset value)))

(defun get-gtk-string (pointer)
  (typecase pointer
    (string pointer)
    (otherwise
     (ukt:trc nil "get-gtk-string sees" pointer (type-of pointer))
     #+allegro (convert-from-cstring pointer)
     #+lispworks (convert-from-foreign-string pointer
                   :null-terminated-p t)
     #-(or allegro lispworks)
     (with-foreign-object (bytes-written :int)
       (g-locale-from-utf8 pointer -1 nil bytes-written nil)))))

(defun to-gtk-string (str)
  "!!!! remember to free returned str pointer"
  (with-foreign-object (bytes-written :int)
    (g-locale-to-utf8 str -1 c-null bytes-written c-null)))

(defmacro with-gdk-threads (&rest body)
  `(unwind-protect
	(progn
	  (gdk-threads-enter)
	  ,@body)
     (gdk-threads-leave)))

(ffx:ff-defun-callable :cdecl :int button-press-event-handler
  ((widget :pointer-void) (signal (* gdk-event-button)) (data :pointer-void))
  (declare (ignorable data))
  (let ((event (gdk-event-button-type signal)))
    (when (or (eql (event-type event) :button_press) 
              (eql (event-type event) :button_release))
      (when (= (gdk-event-button-button signal) 3)
        (gtk-menu-popup widget nil nil nil nil 3 
          (gdk-event-button-time signal)))))
  0)

(defun gtk-widget-set-popup (widget menu)
  (gtk-signal-connect-swap widget "button-press-event"
                           (ffx:ff-register-callable 'button-press-event-handler)
                           :data menu)
  (gtk-signal-connect-swap widget "button-release-event"
                           (ffx:ff-register-callable 'button-press-event-handler)
                           :data menu))

(defun gtk-list-store-new (col-types)
  (let ((c-types (ffx:fgn-alloc :int (length col-types))))
    (loop for type in col-types
          for n upfrom 0
          do (setf (ffx:elti c-types n) (as-gtk-type type)))
    (prog1
        (gtk-list-store-newv (length col-types) c-types)
      (ffx:fgn-free c-types))))


(defun gvi (&optional (key :anon))
  key
;;;  (ukt:trc nil "gvi> " keY)
;;;  (let ((tv (ffx:fgn-alloc :int 32)))
;;;    (dotimes (n 32) (setf (ffx:elti tv n) 0))
;;;    (loop for type in '(:string :icon :int :string)
;;;        do (print (list tv type (as-gtk-type type)
;;;                    (g_value_init tv (as-gtk-type type))))
;;;          (g_value_unset tv))
;;;    (ffx:fgn-free tv))
  )


(defun gtk-list-store-set (lstore iter types-lst data-lst)
  (with-g-value (value)
    (gvi :with-type-val)
    (loop for col from 0
        for data in data-lst
        for type in types-lst       
        for str-ptr = nil #+not (when (or (eql type :string) (eql type :icon))
                                  (gvi :pre-cstring)
                                  (convert-to-cstring data))
        do   
          (gvi :pre-truvi)
          (g-value-init value (as-gtk-type type))
          (funcall (intern (format nil "G-VALUE-SET-~a" (case type 
                                                          (:date 'float)
                                                          (:icon 'string)
                                                          (t type))) 
                     :gtk-ffi)
            value
            (or str-ptr (and (eql type :date) (coerce data 'single-float)) data))
          (gtk-list-store-set-value lstore iter col value)
          (g-value-unset value)
          (when str-ptr (free-cstring str-ptr)))))

(defun gtk-list-store-set-items (store types-lst data-lst)
  (with-tree-iter (iter)
    (dolist (item data-lst)
      (gvi :pre-append)
      (gtk-list-store-append store iter)
      (gvi :pre-set)
      (gtk-list-store-set store iter types-lst item)
      (gvi :post-set))))

(defun gtk-tree-store-new (col-types)
  (let ((gtk-types (ffx:fgn-alloc :int (length col-types))))
    (loop for type in col-types
          for tn upfrom 0
          do (setf (ffx:elti gtk-types tn) (as-gtk-type type)))
    (gtk-tree-store-newv (length col-types) gtk-types)))

(defun gtk-tree-store-set (tstore iter types-lst data-lst)
  "Sets the value of one or more cells in a row referenced by iter."
  (with-g-value (value)
    (loop for col from 0
        for data in data-lst
        for type in types-lst
        do ;; (print (list :tree-store-set value type (as-gtk-type type)))
          (g-value-init value (as-gtk-type type))
          (funcall (case type
		     ((:string :icon) #'g-value-set-string)
		     (:int #'g-value-set-int)
		     (:long #'g-value-set-long)
		     (:boolean #'g-value-set-boolean)
		     ((:float :date) #'g-value-set-float)
		     (t (error "Invalid type: ~S?" type)))
            value
            (if (eql type :date) (coerce data 'single-float) data))
          (gtk-tree-store-set-value tstore iter col value)
          (g-value-unset value))))

;;; todo: The deref-pointer-runtime-typed used by case needs work if
;;; it is going to be used for lispworks, cmu and allegro.
;;; (needs someone who knows how ffi-to-uffi-type maps types for those lisps.)
;;; Even better, eliminate it. It is ill-conceived.
(defun gtk-tree-model-get-cell (model iter column-no cell-type)
  "Returns the item at column-no if column-no [0,<num-columns-1>] or a
   a string like '(0 1 0)', which navigates to the selected item, if 
   column-no = num-columns. (See gtk-tree-store-set-kids)."
  (with-foreign-object (item :pointer-void)
    (gtk-tree-model-get model iter column-no item -1)
  #-(or lispworks cmu allegro) (cast item (as-gtk-type-name cell-type))
  #+allegro
  (case cell-type
    (:string (uffi:convert-from-cstring (uffi:deref-pointer item :cstring)))
    (t (cast item (as-gtk-type-name cell-type))))
  #+lispworks 
   (case cell-type
     (:string (fli:convert-from-foreign-string (deref-pointer item)))
     (t (deref-pointer item)))
   #+cmu
   (case cell-type
     (:string (alien:cast (alien:deref item) c-call:c-string))
     (t (alien:deref item)))))

(defun parse-cell-attrib (attribs)
  (loop for (attrib val) on attribs by #'cddr collect
	(ecase attrib
	  (:foreground (list "foreground" 'c-string val))
	  (:background (list "background" 'c-string val))
	  (:font (list "font" 'c-string val))
	  (:size (list "size-points" 'double-float (coerce val 'double-float)))
	  (:strikethrough (list "strikethrough" 'boolean val)))))

#+cmu
(alien:def-alien-type all-types
  (alien:struct c-struct
    (:string (* t))
    (:icon (* t))
    (:boolean boolean)
    (:int integer)
    (:long c-call:long)
    (:date single-float)
    (:float single-float)
    (:double double-float)))

#-cmu
(progn
  (defun alloc-col-type-buffer (col-type)
    (ecase col-type
      ((:string :icon) (allocate-foreign-object '(:array :cstring) 1))
      (:boolean (allocate-foreign-object '(:array :unsigned-byte) 1)) ;;guess
      (:date (allocate-foreign-object '(:array :float) 1))
      (:int (allocate-foreign-object '(:array :int) 1))
      (:long (allocate-foreign-object '(:array :long) 1))
      (:float (allocate-foreign-object '(:array :float) 1))
      (:double (allocate-foreign-object '(:array :double) 1))))
  
  (defun deref-col-type-buffer (col-type buffer)
    (ecase col-type
      ((:string :icon)
       (get-gtk-string
        (make-pointer (deref-array buffer '(:array :cstring) 0) :cstring)))
      (:boolean (not (zerop (deref-array buffer '(:array :unsigned-byte) 0)))) ;;guess
      (:date (deref-array buffer '(:array :float) 0))
      (:int (deref-array buffer '(:array :int) 0))
      (:long (deref-array buffer '(:array :long) 0))
      (:float (deref-array buffer '(:array :float) 0))
      (:double (deref-array buffer '(:array :double) 0)))))


#+worksforallegroclbutnotlispworks
(defun gtk-tree-view-render-cell (col col-type cell-attrib-f)
  (lambda (tree-column cell-renderer model iter data)
    (DECLARE (ignorable tree-column data))
    (let ((wvar (alloc-col-type-buffer col-type)))
      (gtk-tree-model-get model iter col wvar -1)
      (let ((item-value (deref-col-type-buffer col-type wvar)))
        (with-gtk-string (str (format nil "~a"
                                (if (eql col-type :date)
                                    (multiple-value-bind (sec min hour day month year)
                                        (decode-universal-time (truncate item-value))
                                      (format nil "~2,'0D/~2,'0D/~D ~2,'0D:~2,'0D:~2,'0D"
                                        day month year hour min sec))
                                  item-value)))
          (ukt:trc nil "gtv-render-cell (callback11)> rendering value"
            col col-type item-value)
          (apply #'gtk-object-set-property cell-renderer
            (case col-type
              (:boolean (list "active" 'boolean item-value))
              (:icon (list "stock-id" 'c-string (string-downcase (format nil "gtk-~a" item-value))))
              (t (list "text" 'c-pointer str)))))
        (when cell-attrib-f
          (loop for property in (parse-cell-attrib (funcall cell-attrib-f item-value)) do
                (apply #'gtk-object-set-property cell-renderer property)))
        #-(or allegro lispworks)
        (when (find col-type '(:icon :string))
          (free-foreign-object item-value)))
      (free-foreign-object wvar))))

#-cmu
(defun gtk-tree-view-render-cell (col col-type cell-attrib-f) 
  (ukt:trc nil "gtv-render-cell> creating callback" col col-type cell-attrib-f)
  (lambda (tree-column cell-renderer model iter data)
    (DECLARE (ignorable tree-column data))
    (ukt:trc nil "gtv-render-cell (callback)> entry"
      tree-column cell-renderer model iter data)
    (let ((return-buffer (ffx:fgn-alloc :int 16)))
      (gtk-tree-model-get model iter col
        return-buffer -1)
      (let* ((returned-value (deref-pointer-runtime-typed return-buffer
                               (ffi-to-uffi-type
                                (col-type-to-ffi-type col-type))))
             (ret$ (when (find col-type '(:string :icon))
                     (make-pointer returned-value :cstring)))
             (item-value (cond
                          (ret$ #-lispworks (convert-from-cstring ret$)
                            #+lispworks (convert-from-foreign-string ret$
                                          :null-terminated-p t))
                          ((eq col-type :boolean)
                           (not (zerop returned-value)))
                          (t returned-value))))
        (ukt:trc nil "gtv-render-cell (callback)>> rendering value"
          col col-type ret$ item-value)

        (apply #'gtk-object-set-property cell-renderer 
          (case col-type 
            (:boolean (list "active" 'boolean item-value))
            (:icon (list "stock-id" 'c-string
                     (string-downcase (format nil "gtk-~a" item-value))))
            (t (list "text" 'c-string
                 (case col-type
                   (:date (multiple-value-bind (sec min hour day month year) 
                              (decode-universal-time (truncate item-value))
                            (format nil "~2,'0D/~2,'0D/~D ~2,'0D:~2,'0D:~2,'0D" 
                              day month year hour min sec)))
                   (:string (get-gtk-string item-value))
                   (otherwise (format nil "~a" item-value)))))))

        (when cell-attrib-f 
          (loop for property in (parse-cell-attrib (funcall cell-attrib-f item-value))
              do (apply #'gtk-object-set-property cell-renderer property)))
        (when ret$
          (uffi:free-foreign-object ret$))
        (ffx:fgn-free return-buffer)))
    1))

#+cmu
(defun gtk-tree-view-render-cell (col col-type cell-attrib-f)
    #'(lambda (tree-column cell-renderer model iter data)
        (alien:with-alien ((struct all-types))
          (gtk-tree-model-get model iter col
            (alien:addr (alien:slot struct col-type))
            -1)
          (let ((item-value (if (or (eql col-type :string) (eql col-type :icon))
                                (get-gtk-string (alien:slot struct col-type))
                              (alien:slot struct col-type))))
            (with-gtk-string (str (format nil "~a"
                                    (if (eql col-type :date)
                        (multiple-value-bind (sec min hour day month year)
                        (decode-universal-time (truncate item-value))
                          (format nil "~2,'0D/~2,'0D/~D ~2,'0D:~2,'0D:~2,'0D"
                              day month year hour min sec))
                        item-value)))
        (ukt:trc nil "gtv-render-cell (callback11)> rendering value"
             col col-type item-value)
        (apply #'gtk-object-set-property cell-renderer
           (case col-type
             (:boolean (list "active" 'boolean item-value))
             (:icon (list "stock-id" 'c-string (string-downcase (format nil "gtk-~a" item-value))))
             (t (list "text" 'c-pointer str)))))
      (when cell-attrib-f
        (loop for property in (parse-cell-attrib (funcall cell-attrib-f item-value)) do
         (apply #'gtk-object-set-property cell-renderer property))))
    (when (eql col-type :string)
      (g-free (alien:slot struct :string))))))


#+clisp

(defun gtk-tree-view-render-cell (col col-type cell-attrib-f) 
  #'(lambda (tree-column cell-renderer model iter data)
      (declare (ignore data))
      (with-c-var 
          (struct '(c-struct list
                     (:string c-pointer)
                     (:icon c-pointer)
                     (:boolean boolean)
                     (:int int)
                     (:long long)
                     (:date single-float)
                     (:float single-float)
                     (:double double-float))
            (list nil nil nil 0 0 (coerce 0 'single-float) (coerce 0 'single-float) (coerce 0 'double-float)))
        (gtk-tree-model-get model iter col
          (c-var-address (slot struct col-type))
          -1)
        (let ((item-value (if (or (eql col-type :string) (eql col-type :icon))
                              (get-gtk-string (slot struct col-type))
                            (slot struct col-type))))
          (ukt:trc nil "tv-render-cell: "
            :col-type col-type 
            :item item-value)
          (with-gtk-string (str (format nil "~a" 
                                  (if (eql col-type :date)
                                      (multiple-value-bind (sec min hour day month year) 
                                          (decode-universal-time (truncate item-value))
                                        (format nil "~2,'0D/~2,'0D/~D ~2,'0D:~2,'0D:~2,'0D" 
                                          day month year hour min sec))
                                    item-value)))
            (apply #'gtk-object-set-property cell-renderer 
              (case col-type 
                (:boolean (list "active" 'boolean item-value))
                (:icon (list "stock-id" 'c-string (string-downcase (format nil "gtk-~a" item-value))))
                (t (list "text" 'c-pointer str)))))
          (when cell-attrib-f 
            (loop for property in (parse-cell-attrib (funcall cell-attrib-f item-value)) do
                  (apply #'gtk-object-set-property cell-renderer property))))
        (when (eql col-type :string)
          (g-free (slot struct :string))))))

(defun gtk-file-chooser-get-filenames-strs (file-chooser)
  (let ((glist (gtk-file-chooser-get-filenames file-chooser)))
    (loop for lst-address = glist then (get-slot-value lst-address 'gslist 'next)
        while (and lst-address (not (zerop lst-address)))
        collect (get-slot-value lst-address 'gslist 'data)
        finally (g-slist-free glist))))

(eval-when (compile load eval)
  (export '(gtk-signal-connect gtk-signal-connect-swap gtk-object-set-property
             with-gtk-string get-gtk-string to-gtk-string 
             with-gdk-threads make-gtk-tree-iter with-tree-iter
             gtk-widget-set-popup gvi
             gtk-list-store-new gtk-list-store-set gtk-list-store-set-items
             gtk-tree-store-new gtk-tree-store-set gtk-tree-store-set-kids
             gtk-tree-model-get-cell
             gtk-tree-view-render-cell
             gtk-file-chooser-get-filenames-strs
             )))

